from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

from opencosmo import collection
from opencosmo.header import read_header
from opencosmo.units.convention import UnitConvention

if TYPE_CHECKING:
    import h5py

    from opencosmo.header import OpenCosmoHeader


class FILE_TYPE(Enum):
    HALO_PROPERTIES = 0
    HALO_PARTICLES = 1
    GALAXY_PROPERTIES = 2
    GALAXY_PARTICLES = 3
    SOD_BINS = 4
    LIGHTCONE = 5
    STRUCTURE_COLLECTION = 6
    SIMULATION_COLLECTION = 7
    SYNTHETIC_CATALOG = 8
    HEALPIX_MAP = 9


def get_file_type(file: h5py.File) -> FILE_TYPE:
    if "header" in file.keys():
        dtype = file["header"]["file"].attrs["data_type"]
        if dtype == "halo_particles":
            return FILE_TYPE.HALO_PARTICLES
        elif dtype == "halo_profiles":
            return FILE_TYPE.SOD_BINS
        elif dtype == "halo_properties":
            return FILE_TYPE.HALO_PROPERTIES
        elif dtype == "galaxy_properties":
            return FILE_TYPE.GALAXY_PROPERTIES
        elif dtype == "galaxy_particles":
            return FILE_TYPE.GALAXY_PARTICLES
        elif dtype == "diffsky_fits":
            return FILE_TYPE.SYNTHETIC_CATALOG
        elif dtype == "healpix_map":
            return FILE_TYPE.HEALPIX_MAP
        else:
            raise ValueError(f"Unknown file type {dtype}")

    if not all("header" in group.keys() for group in file.values()):
        for subgroup in file.values():
            if not all("header" in g.keys() for g in subgroup.values()):
                raise ValueError(
                    "Unknown file type. "
                    "It appears to have multiple datasets, but organized incorrectly"
                )
    if all(group["header"]["file"].attrs["is_lightcone"] for group in file.values()):
        return FILE_TYPE.LIGHTCONE
    elif (
        len(set(group["header"]["file"].attrs["data_type"] for group in file.values()))
        == 1
    ):
        return FILE_TYPE.SIMULATION_COLLECTION

    elif all("data" not in group.keys() for group in file.values()):
        for group in file.values():
            sub_groups = {
                g["header"]["file"].attrs["data_type"]: g for g in group.values()
            }
            collection.structure.io.validate_linked_groups(sub_groups)
        return FILE_TYPE.SIMULATION_COLLECTION
    else:
        group = {name: group for name, group in file.items()}
        collection.structure.io.validate_linked_groups(group)
        return FILE_TYPE.STRUCTURE_COLLECTION


class OpenTarget:
    def __init__(self, group: h5py.Group | h5py.File, header: OpenCosmoHeader):
        self.group = group
        self.header = header

    @property
    def data_type(self):
        return self.header.file.data_type


def make_all_targets(files: list[h5py.File]):
    bad_files = []
    targets = []
    for file in files:
        try:
            targets += make_file_targets(file)
        except ValueError:
            bad_files.append(file.filename)
            raise
    if bad_files:
        raise ValueError(
            f"Some files were not able to be opened. They may not be OpenCosmo files: {bad_files}"
        )

    return targets


def make_file_targets(file: h5py.File):
    try:
        header = read_header(file, unit_convention=UnitConvention.COMOVING)
    except KeyError:
        header = None
    if header is not None and "data" in file.keys():
        return [OpenTarget(file, header)]
    if header is None and "data" in file.keys():
        raise ValueError(
            f"The file at {file.file.filename} appears to be missing a header. "
            "Are you sure it is an OpenCosmo file?"
        )
    if header is None:
        headers = {name: read_header(group) for name, group in file.items()}
    else:
        headers = {name: header for name in file.keys() if name != "header"}

    output = []
    for name, header in headers.items():
        target = OpenTarget(file[name], header)
        output.append(target)
    return output


def evaluate_load_conditions(targets: list[OpenTarget], open_kwargs: dict[str, bool]):
    """
    Datasets can define conditional loading via an addition group called "load/if".
    the "if" group can define parameters which must either be true or false for the
    given group to be loaded. These parameters can then be provided by the user to the
    "open" function. Parameters not specified by the user default to False.

    Note that some open kwargs may be used in other places in the opening process,
    and will just be ignored here.
    """
    if len(targets) == 1:
        return targets
    output = []
    for target in targets:
        try:
            ifgroup = target.group["load/if"]
        except KeyError:
            output.append(target)
            continue
        load = True
        for key, condition in ifgroup.attrs.items():
            load = load and (open_kwargs.get(key, False) == condition)
        if load:
            output.append(target)
    return output



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Any, Callable

import astropy.units as u
import numpy as np

from opencosmo.evaluate import insert_data, prepare_kwargs

if TYPE_CHECKING:
    from opencosmo import Dataset
    from opencosmo.index import ChunkedIndex


class EvaluateStrategy(Enum):
    VECTORIZE = "vectorize"
    ROW_WISE = "row_wise"
    CHUNKED = "chunked"


def evaluate_rows(data: dict[str, np.ndarray], func: Callable, kwargs: dict[str, Any]):
    data_length = len(next(iter(data.values())))
    kwargs_, iterable_kwargs = prepare_kwargs(data_length, kwargs)
    iterable_data = data | iterable_kwargs
    storage = {}
    for i in range(data_length):
        iterable_inputs = {name: values[i] for name, values in iterable_data.items()}
        output = func(**iterable_inputs, **kwargs_)
        if not isinstance(output, dict):
            output = {func.__name__: output}
        if i == 0:
            storage = __make_row_based_output_from_first_values(output, data_length)
            continue
        insert_data(storage, i, output)
    return storage


def __make_row_based_output_from_first_values(values, data_length):
    storage = {}
    for name, value in values.items():
        try:
            shape = (data_length,) + value.shape
        except AttributeError:
            shape = (data_length,)
        try:
            dtype = value.dtype
        except AttributeError:
            dtype = type(value)
        column_storage = np.zeros(shape, dtype=dtype)
        if isinstance(value, u.Quantity):
            column_storage *= value.unit
        column_storage[0] = value
        storage[name] = column_storage

    return storage


def evaluate_chunks(
    data: dict[str, np.ndarray],
    func: Callable,
    kwargs: dict[str, Any],
    index: ChunkedIndex,
):
    data_length = len(next(iter(data.values())))
    kwargs_, iterable_kwargs = prepare_kwargs(data_length, kwargs)
    input_data = data | iterable_kwargs

    chunk_splits = np.cumsum(index[1])
    storage = {}
    input_data = {name: np.split(arr, chunk_splits) for name, arr in data.items()}
    for i in range(len(chunk_splits)):
        chunk_input_data = {name: split[i] for name, split in input_data.items()}
        output = func(**chunk_input_data, **kwargs_)
        if not isinstance(output, dict):
            output = {func.__name__: output}
        if i == 0:
            storage = __make_chunked_based_output_from_first_values(output, data_length)
            continue
        for name, values in output.items():
            storage[name][chunk_splits[i - 1] : chunk_splits[i]] = values
    return storage


def __make_chunked_based_output_from_first_values(values, data_length):
    storage = {}
    for name, value in values.items():
        shape = (data_length,) + value.shape[1:]
        dtype = value.dtype
        column_storage = np.zeros(shape, dtype=dtype)
        if isinstance(value, u.Quantity):
            column_storage *= value.unit
        column_storage[0 : len(value)] = value
        storage[name] = column_storage

    return storage


def evaluate_vectorized(data, func, kwargs):
    return func(**data, **kwargs)


def do_first_evaluation(
    func: Callable, strategy: str, format: str, kwargs: dict[str, Any], dataset: Dataset
):
    eval_strategy = EvaluateStrategy(strategy)
    match eval_strategy:
        case EvaluateStrategy.VECTORIZE:
            values = dataset.take(1).get_data(format, unpack=False)
            try:
                values = dict(values)
            except TypeError:
                values = {dataset.columns[0]: values}

            return func(**values, **kwargs), eval_strategy

        case EvaluateStrategy.ROW_WISE:
            values = dataset.take(1).get_data(format, unpack=True)
            try:
                values = dict(values)
            except TypeError:
                values = {dataset.columns[0]: values}
            return func(**values, **kwargs), eval_strategy

        case EvaluateStrategy.CHUNKED:
            index = dataset.index
            assert isinstance(index, tuple)
            first_chunk_size = index[1][0]
            first_chunk = dataset.take(first_chunk_size, at="start").get_data(format)
            first_chunk = dict(first_chunk)
            return func(**first_chunk, **kwargs), eval_strategy



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Callable, Optional

import h5py

import opencosmo as oc
from opencosmo import collection
from opencosmo.dataset import state as dss
from opencosmo.dataset.handler import Hdf5Handler
from opencosmo.file import FileExistance, resolve_path
from opencosmo.index.build import empty, from_range
from opencosmo.io.file import (
    evaluate_load_conditions,
    get_file_type,
    make_all_targets,
)
from opencosmo.io.serial import allocate, write_columns, write_metadata
from opencosmo.mpi import get_comm_world
from opencosmo.spatial.builders import from_model
from opencosmo.spatial.region import FullSkyRegion
from opencosmo.spatial.tree import open_tree
from opencosmo.units import UnitConvention

if TYPE_CHECKING:
    from pathlib import Path
    from types import ModuleType

    from opencosmo.index import ChunkedIndex
    from opencosmo.io.file import OpenTarget

    from .protocols import Writeable

    mpiio: Optional[ModuleType]
    partition: Optional[Callable]

if get_comm_world() is not None:
    from opencosmo.dataset.mpi import partition
    from opencosmo.io import mpi as mpiio
else:
    mpiio = None
    partition = None

    """
    This module defines the main user-facing io functions: open and write

    open can take any number of file paths, and will always construct a single object 
    (either a dataset or a collection).

    write takes exactly one path and exactly one opencosmo dataset or collection

    open works in the following way:

    1. Read headers and get dataset names and types for all files passed
    2. If there is only a single dataset, simply open it as such
    3. If there are multiple datasets, user the headers to determine
       if the dataset are compatible (i.e. capabale of existing together in
       a collection)
    4. Open all datasets individually
    5. Call the merge functionality for the appropriate collection.
    """


class COLLECTION_TYPE(Enum):
    LIGHTCONE = 0
    STRUCTURE_COLLECTION = 1
    SIMULATION_COLLECTION = 2


def open(
    *files: str | Path | h5py.File | h5py.Group, **open_kwargs: bool
) -> oc.Dataset | collection.Collection:
    """
    Open a dataset or data collection from one or more opencosmo files.

    If you open a file with this function, you should generally close it
    when you're done

    .. code-block:: python

        import opencosmo as oc
        ds = oc.open("path/to/file.hdf5")
        # do work
        ds.close()

    Alternatively you can use a context manager, which will close the file
    automatically when you are done with it.

    .. code-block:: python

        import opencosmo as oc
        with oc.open("path/to/file.hdf5") as ds:
            # do work

    When you have multiple files that can be combined into a collection,
    you can use the following.

    .. code-block:: python

        import opencosmo as oc
        ds = oc.open("haloproperties.hdf5", "haloparticles.hdf5")


    Parameters
    ----------
    *files: str or pathlib.Path
        The path(s) to the file(s) to open.

    **open_kwargs: bool
        True/False flags that can be used to only load certain datasets from
        the files. Check the documentation for the data type you are working
        with for available flags. Will be ignored if only one file is passed
        and the file only contains a single dataset.

    Returns
    -------
    dataset : oc.Dataset or oc.Collection
        The dataset or collection opened from the file.

    """
    if len(files) == 1 and isinstance(files[0], list):
        file_list = files[0]
    else:
        file_list = list(files)
    file_list.sort()
    handles = [h5py.File(f) for f in file_list]
    targets = make_all_targets(handles)
    targets = evaluate_load_conditions(targets, open_kwargs)
    file_types = list(map(get_file_type, handles))
    if len(targets) > 1:
        collection_type = collection.get_collection_type(targets, file_types)
        return collection_type.open(targets, **open_kwargs)

    else:
        return open_single_dataset(targets[0])

    # For now the only way to open multiple files is with a StructureCollection


def open_single_dataset(
    target: OpenTarget,
    metadata_group: Optional[str] = None,
    bypass_lightcone: bool = False,
    bypass_mpi: bool = False,
):
    header = target.header
    handle = target.group

    assert header is not None

    try:
        tree = open_tree(
            handle,
            header.with_units("scalefree").simulation["box_size"].value,
            header.file.is_lightcone,
        )
    except ValueError:
        tree = None

    if header.file.region is not None:
        sim_region = from_model(header.file.region)
    elif header.file.is_lightcone:
        sim_region = FullSkyRegion()
    else:
        p1 = (0, 0, 0)
        p2 = tuple(header.simulation["box_size"].value for _ in range(3))
        sim_region = oc.make_box(p1, p2)

    index: Optional[ChunkedIndex] = None
    handler = Hdf5Handler.from_group(handle["data"])

    if not bypass_mpi and (comm := get_comm_world()) is not None:
        assert partition is not None
        try:
            idx_data = handle["index"]
            part = partition(comm, len(handler), idx_data, tree)
            if part is None:
                index = empty()
            else:
                index = part.idx
                sim_region = part.region if part.region is not None else sim_region
        except KeyError:
            n_ranks = comm.Get_size()
            n_per = len(handler) // n_ranks
            chunk_boundaries = [i * n_per for i in range(n_ranks + 1)]
            chunk_boundaries[-1] = len(handler)
            rank = comm.Get_rank()
            index = from_range(chunk_boundaries[rank], chunk_boundaries[rank + 1])

    if metadata_group is not None:
        metadata_group = handle[metadata_group]

    elif "metadata" in handle.keys():
        metadata_group = handle["metadata"]

    state = dss.DatasetState.from_group(
        handle["data"],
        header,
        UnitConvention.COMOVING,
        sim_region,
        index,
        metadata_group,
    )

    dataset = oc.Dataset(
        header,
        state,
        tree=tree,
    )
    if header.file.data_type == "healpix_map":
        return collection.HealpixMap(
            {"data": dataset},
            header.healpix_map["nside"],
            header.healpix_map["nside_lr"],
            header.healpix_map["ordering"],
            header.healpix_map["full_sky"],
            header.healpix_map["z_range"],
        )
    elif header.file.is_lightcone and not bypass_lightcone:
        return collection.Lightcone({"data": dataset}, header.lightcone["z_range"])

    return dataset


def write(path: Path, dataset: Writeable, overwrite=False) -> None:
    """
    Write a dataset or collection to the file at the sepecified path.

    Parameters
    ----------
    file : str or pathlib.Path
        The path to the file to write to.
    dataset : oc.Dataset
        The dataset to write.
    overwrite : bool, default = False
        If the file already exists, overwrite it


    Raises
    ------
    FileExistsError
        If the file at the specified path already exists and overwrite is False
    FileNotFoundError
        If the parent folder of the ouput file does not exist
    """

    existance_requirement = FileExistance.MUST_NOT_EXIST
    if overwrite:
        existance_requirement = FileExistance.EITHER

    path = resolve_path(path, existance_requirement)

    schema = dataset.make_schema()

    if mpiio is not None:
        return mpiio.write_parallel(path, schema)

    file = h5py.File(path, "w")
    allocate(file, schema)
    write_metadata(file, schema)
    write_columns(file, schema)



import astropy.cosmology.units as cu  # type: ignore
import astropy.units as u  # type: ignore

from .convention import UnitConvention

_ = u.add_enabled_units(cu)


class UnitsError(Exception):
    pass


__all__ = ["UnitConvention", "UnitsError"]



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Optional

import numpy as np

from opencosmo import io
from opencosmo.collection import lightcone as lc
from opencosmo.collection.structure import structure as sc

if TYPE_CHECKING:
    import h5py

    from opencosmo import dataset as d
    from opencosmo.header import OpenCosmoHeader

ALLOWED_LINKS = {  # h5py.Files that can serve as a link holder and
    "halo_properties": ["halo_particles", "halo_profiles", "galaxy_properties"],
    "galaxy_properties": ["galaxy_particles"],
}


def remove_empty(dataset):
    metadata = dataset.get_metadata()
    mask = np.ones(len(dataset), dtype=bool)
    for name, col in metadata.items():
        if "size" in name:
            mask &= col != 0
        elif "idx" in name:
            mask &= col != -1

    if not mask.all():
        dataset = dataset.take_rows(np.where(mask)[0])
    return dataset


def validate_linked_groups(groups: dict[str, h5py.Group]):
    if "halo_properties" in groups:
        if "data_linked" not in groups["halo_properties"].keys():
            raise ValueError(
                "File appears to be a structure collection, but does not have links!"
            )
    elif "galaxy_properties" in groups:
        if "data_linked" not in groups["galaxy_properties"].keys():
            raise ValueError(
                "File appears to be a structure collection, but does not have links!"
            )
    if len(groups) == 1:
        raise ValueError("Structure collections must have more than one dataset")


def get_linked_datasets(
    linked_files_by_type: dict[str, h5py.File | h5py.Group],
    header: OpenCosmoHeader,
):
    targets = {}
    for dtype, pointer in linked_files_by_type.items():
        if "data" not in pointer.keys():
            targets.update(
                {
                    k: io.io.OpenTarget(pointer[k], header)
                    for k in pointer.keys()
                    if k != "header"
                }
            )
        else:
            targets.update({dtype: io.io.OpenTarget(pointer, header)})
    datasets = {
        dtype: io.io.open_single_dataset(target, bypass_lightcone=True, bypass_mpi=True)
        for dtype, target in targets.items()
    }
    return datasets


def build_structure_collection(targets: list[io.io.OpenTarget], ignore_empty: bool):
    link_sources = defaultdict(list)
    link_targets: dict[str, dict[str, list[d.Dataset | sc.StructureCollection]]] = (
        defaultdict(lambda: defaultdict(list))
    )
    for target in targets:
        if target.data_type == "halo_properties":
            link_sources["halo_properties"].append(target)
        elif target.data_type == "galaxy_properties":
            link_sources["galaxy_properties"].append(target)
        elif target.data_type.startswith("halo"):
            dataset = io.io.open_single_dataset(
                target, bypass_lightcone=True, bypass_mpi=True
            )
            name = target.group.name.split("/")[-1]
            if not name:
                name = target.data_type
            elif name.startswith("halo_properties"):
                name = name[16:]
            link_targets["halo_targets"][name].append(dataset)
        elif target.data_type.startswith("galaxy"):
            dataset = io.io.open_single_dataset(
                target, bypass_lightcone=True, bypass_mpi=True
            )
            name = target.group.name.split("/")[-1]
            if not name:
                name = target.data_type
            elif name.startswith("galaxy_properties"):
                name = name[18:]
            link_targets["galaxy_targets"][name].append(dataset)
        else:
            raise ValueError(
                f"Unknown data type for structure collection {target.data_type}"
            )

    if (
        len(link_sources["halo_properties"]) > 1
        or len(link_sources["galaxy_properties"]) > 1
    ):
        raise NotImplementedError(
            "Opening structure collections that span multiple redshifts is not currently supported"
        )
        # Potentially a lightcone structure collection
        collections = {}
        sources_by_step, targets_by_step = __sort_by_step(link_sources, link_targets)
        if set(sources_by_step.keys()) != set(targets_by_step.keys()):
            raise ValueError("Datasets are not the same across all lightcone steps!")
        for step, sources in sources_by_step.items():
            halo_properties = sources.get("halo_properties")
            galaxy_properties = sources.get("galaxy_properties")
            targets = targets_by_step[step]
            collection = __build_structure_collection(
                halo_properties, galaxy_properties, targets, ignore_empty
            )
            collections[step] = collection

        expected_datasets = set(next(iter(collections.values())).keys())
        for collection in collections.values():
            if set(collection.keys()) != expected_datasets:
                raise ValueError(
                    "All structure collections in a lightcone must have the same set of datasets"
                )
        return lc.Lightcone(collections)

    halo_properties_target = None
    galaxy_properties_target = None
    if link_sources["halo_properties"]:
        halo_properties_target = link_sources["halo_properties"][0]
    if link_sources["galaxy_properties"]:
        galaxy_properties_target = link_sources["galaxy_properties"][0]

    input_link_targets: dict[str, dict[str, d.Dataset | sc.StructureCollection]] = (
        defaultdict(dict)
    )
    for source_type, source_targets in link_targets.items():
        if any(len(ts) > 1 for ts in source_targets.values()):
            raise ValueError("Found more than one linked file of a given type!")
        input_link_targets[source_type] = {
            key: t[0] for key, t in source_targets.items()
        }

    return __build_structure_collection(
        halo_properties_target,
        galaxy_properties_target,
        input_link_targets,
        ignore_empty,
    )


def __sort_by_step(link_sources: dict[str, list[io.io.OpenTarget]], link_targets):
    sources_by_step: dict[int, dict[str, io.io.OpenTarget]] = defaultdict(dict)
    targets_by_step: dict[int, dict[str, dict[str, d.Dataset]]] = defaultdict(
        lambda: defaultdict(dict)
    )
    for source_name, sources in link_sources.items():
        for source in sources:
            if not source.header.file.is_lightcone:
                raise ValueError(
                    "Recived multiple source datasets of a single type, but not all are lightcone datasets!"
                )
            if source.header.file.step is None:
                raise ValueError("No step in source!")

            sources_by_step[source.header.file.step][source_name] = source
    for target_type, targets_ in link_targets.items():
        for target_name, targets in targets_.items():
            for target in targets:
                if not target.header.file.is_lightcone:
                    raise ValueError(
                        "Recived multiple datasets of a single type, but not all are lightcone datasets!"
                    )
                targets_by_step[target.header.file.step][target_type][target_name] = (
                    target
                )

    return sources_by_step, targets_by_step


def __build_structure_collection(
    halo_properties_target: Optional[io.io.OpenTarget],
    galaxy_properties_target: Optional[io.io.OpenTarget],
    link_targets: dict[str, dict[str, d.Dataset | sc.StructureCollection]],
    ignore_empty: bool,
):
    if galaxy_properties_target is not None and "galaxy_targets" in link_targets:
        source_dataset = io.io.open_single_dataset(
            galaxy_properties_target,
            metadata_group="data_linked",
            bypass_lightcone=True,
            bypass_mpi=halo_properties_target is not None,
        )
        if ignore_empty and halo_properties_target is None:
            source_dataset = remove_empty(source_dataset)
        collection = sc.StructureCollection(
            source_dataset,
            source_dataset.header,
            link_targets["galaxy_targets"],
        )
        if halo_properties_target is not None:
            link_targets["halo_targets"]["galaxy_properties"] = collection
        else:
            return collection

    if (
        halo_properties_target is not None
        and galaxy_properties_target is not None
        and "galaxy_targets" not in link_targets
    ):
        galaxy_properties = io.io.open_single_dataset(
            galaxy_properties_target, bypass_lightcone=True
        )
        link_targets["halo_targets"]["galaxy_properties"] = galaxy_properties

    if halo_properties_target is not None and link_targets["halo_targets"]:
        source_dataset = io.io.open_single_dataset(
            halo_properties_target, metadata_group="data_linked", bypass_lightcone=True
        )
        if ignore_empty:
            source_dataset = remove_empty(source_dataset)

        return sc.StructureCollection(
            source_dataset,
            source_dataset.header,
            link_targets["halo_targets"],
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from .schema import FileEntry

if TYPE_CHECKING:
    from .schema import Schema


""" 
Verification is just a check to make sure that the data going into the file is valid. It is intentionally centralized,
such that if you create a new data type and implement "make_schema" it will fail at this step until you add verification.
"""


class ZeroLengthError(Exception):
    pass


def verify_file(
    schema: Schema,
):
    match schema.type:
        case FileEntry.DATASET:
            return verify_dataset_data(schema)
        case FileEntry.STRUCTURE_COLLECTION:
            return verify_structure_collection_data(schema)
        case FileEntry.LIGHTCONE:
            verify_lightcone_collection_schema(schema)
        case FileEntry.SIMULATION_COLLECTION:
            for name, ds_schema in schema.children.items():
                verify_dataset_data(ds_schema)
        case FileEntry.HEALPIX_MAP:
            verify_dataset_data(schema, has_index=False)
        case _:
            raise ValueError("Unknown file structure!")


def verify_column_group(schema: Schema, require_data: bool = False):
    """
    Verify that a given data group is valid. This requires that:
    1. All column writers have the same length
    2. All columns have the same combine strategy
    3. All columns are in the same group
    """
    column_names = set()
    group_names = set()
    column_lengths = {}
    column_strategies = set()
    for column_path, column_writer in schema.columns.items():
        try:
            group_name, column_name = column_path.rsplit("/", 1)
        except ValueError:
            group_name = ""
            column_name = column_path
        group_names.add(group_name)
        column_names.add(column_name)
        column_lengths[column_path] = len(column_writer)
        column_strategies.add(column_writer.combine_strategy)

    all_column_lengths = set(column_lengths.values())

    if len(all_column_lengths) != 1:
        raise ValueError(
            "Columns within a single group should always have the same length!"
        )
    elif (group_length := all_column_lengths.pop()) == 0 and require_data:
        raise ZeroLengthError

    if len(column_strategies) != 1:
        raise ValueError(
            "Columns within a single group should always have the same combine strategy!"
        )

    return (group_names.pop(), group_length, column_strategies.pop())


def verify_dataset_data(schema: Schema, has_index=True):
    """
    Verify a given dataset is valid. Requiring:
    1. It has a data group
    2. It has a spatial index group (if has_index = True)
    3. If it has any metadata groups, they are the same length as the data group

    Once this is verified, we delegate to verify_column_group to ensure
    individual column groups are valid.
    """
    children = schema.children

    if "data" not in children or ("index" not in children and has_index):
        raise ValueError("Datasets must have at least a data group and a index group")

    metadata_groups = [
        child
        for name, child in schema.children.items()
        if name not in ["data", "index"] and child.type == FileEntry.COLUMNS
    ]

    _, data_length, data_combine_strategy = verify_column_group(
        schema.children["data"], require_data=True
    )
    if has_index:
        for child in schema.children["index"].children.values():
            verify_column_group(child)
    for md_child in metadata_groups:
        _, md_length, md_combine_strategy = verify_column_group(md_child)
        if md_length != data_length or md_combine_strategy != data_combine_strategy:
            raise ValueError(
                "Metadata groups must be the same length and have the same combine strategy as data groups!"
            )


def verify_lightcone_collection_schema(schema: Schema):
    """
    Verify a lightcone collection. Note that a single dataset
    can also technically be a lighcone collection, if is_lightcone is
    set to true in its header. Mostly just delegates to underlying
    dataset checks.
    """
    if len(schema.children) < 1:
        raise ValueError("Expect at least one lightcone child!")
    elif "data" in schema.children:
        # Single-dataset lightcone
        return verify_dataset_data(schema)
    for key, child_schema in schema.children.items():
        verify_dataset_data(child_schema)


def verify_structure_collection_data(schema: Schema):
    """
    Structure collections have a lot going on, but they are mostly just datasets. The only
    thing we have to check is that we have an explicit "data_linked" group in any link
    holders, and we then verify the individual dataset.

    """
    if "halo_properties" in schema.children:
        link_holder = "halo_properties"
    elif "galaxy_properties" in schema.children:
        link_holder = "galaxy_properties"
    else:
        raise ValueError("No valid link holder found in schema!")

    for child_name, child_schema in schema.children.items():
        if child_name == link_holder:
            has_link = any(
                map(lambda cn: "data_linked" in cn, child_schema.children.keys())
            )
            if not has_link:
                raise ValueError(
                    f'Source dataset {child_name} does not have expected "data_linked" group'
                )

        match child_schema.type:
            case FileEntry.DATASET:
                verify_dataset_data(child_schema)
            case FileEntry.STRUCTURE_COLLECTION:
                verify_structure_collection_data(child_schema)
            case _:
                raise ValueError("Got an unknown child for structure collection!")



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import astropy.units as u
import numpy as np

if TYPE_CHECKING:
    from opencosmo.units.handler import UnitHandler


def resort(columns: dict[str, np.ndarray], sorted_index: Optional[np.ndarray]):
    if sorted_index is None or not columns:
        return columns
    reverse_sort = np.argsort(sorted_index)
    return {name: data[reverse_sort] for name, data in columns.items()}


def validate_in_memory_columns(
    columns: dict[str, np.ndarray], unit_handler: UnitHandler, ds_length: int
):
    new_units = {}
    for colname, column in columns.items():
        if len(column) != ds_length:
            raise ValueError(f"Column {colname} is not the same length as the dataset!")
        if isinstance(column, u.Quantity):
            new_units[colname] = column.unit
        else:
            new_units[colname] = None

    return unit_handler.with_static_columns(**new_units)



from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, Iterable, Optional

import healpy as hp
import numpy as np

from opencosmo import dataset as ds
from opencosmo.io.schema import FileEntry, make_schema
from opencosmo.mpi import get_comm_world
from opencosmo.spatial.check import find_coordinates_2d

if TYPE_CHECKING:
    from opencosmo.io.schema import Schema
    from opencosmo.mpi import MPI


def update_order(data: np.ndarray, comm: Optional[MPI.Comm], order: np.ndarray):
    return data[order]


def sync_headers(datasets: list[ds.Dataset], redshift_range):
    steps = (
        dataset.header.file.step
        for dataset in datasets
        if dataset.header.file.step is not None
    )
    redshifts = (
        dataset.header.file.redshift
        for dataset in datasets
        if dataset.header.file.redshift is not None
    )
    step = max(steps)
    redshift = max(redshifts)

    if (comm := get_comm_world()) is not None:
        step = np.max(comm.allgather(step))
        redshift = max(comm.allgather(redshift))
        z_ranges = comm.allgather(redshift_range)
        z_min = min(zr[0] for zr in z_ranges)
        z_max = max(zr[1] for zr in z_ranges)
        redshift_range = (z_min, z_max)

    # lightcones are identified by their upper redshift slice
    header_schema = datasets[0].header.dump()
    header_schema.attributes["file"]["redshift"] = redshift
    header_schema.attributes["file"]["step"] = step
    header_schema.attributes["lightcone"]["z_range"] = redshift_range
    return header_schema


def stack_lightcone_datasets_in_schema(
    datasets: list[ds.Dataset], name: str, redshift_range: tuple[float, float]
):
    if len(datasets) == 1 and get_comm_world() is None:
        schema = datasets[0].make_schema(name=name)
        header = sync_headers(datasets, redshift_range)
        schema.children["header"] = header
        return schema

    schemas = [ds.make_schema(name=name) for ds in datasets]
    index_names = list(schemas[0].children["index"].children.keys())
    index_names.sort()
    max_level = int(index_names[-1][-1])

    assert all(isinstance(dataset, ds.Dataset) for dataset in datasets)
    new_data_group = stack_data_groups([schema.children["data"] for schema in schemas])

    order = get_stacked_lightcone_order(datasets, max_level)
    updater = partial(update_order, order=order)

    for column in new_data_group.columns.values():
        column.set_transformation(updater)

    new_index_group = stack_index_groups(
        [schema.children["index"] for schema in schemas]
    )
    header_schema = sync_headers(datasets, redshift_range)

    children = {
        "data": new_data_group,
        "index": new_index_group,
        "header": header_schema,
    }

    return make_schema(name, FileEntry.DATASET, children=children)


def stack_index_groups(schemas: list[Schema]):
    base_schema = schemas[0]
    new_children = {}
    for index_level in base_schema.children.keys():
        all_level_schemas = [schema.children[index_level] for schema in schemas]
        new_children[index_level] = stack_data_groups(all_level_schemas)
    return make_schema("index", FileEntry.COLUMNS, new_children)


def stack_data_groups(schemas: list[Schema]):
    if len(schemas) == 1:
        return schemas[0]
    base_schema = schemas[0]
    new_writers = {}
    for name, column_writer in base_schema.columns.items():
        other_writers = [schema.columns[name] for schema in schemas[1:]]
        new_writer = column_writer.combine(other_writers)
        new_writers[name] = new_writer

    new_schema = make_schema(
        base_schema.name,
        base_schema.type,
        children={},
        columns=new_writers,
        attributes=base_schema.attributes,
    )
    return new_schema


def get_stacked_lightcone_order(datasets: Iterable[ds.Dataset], max_index_depth: int):
    datasets = list(datasets)
    nside = 2**max_index_depth
    coordinates = list(map(find_coordinates_2d, datasets))
    pixels = np.concatenate(
        [
            hp.ang2pix(nside, coords.ra.value, coords.dec.value, lonlat=True, nest=True)
            for coords in coordinates
        ]
    )
    return np.argsort(pixels)



from __future__ import annotations

import numba as nb  # type: ignore
import numpy as np

SimpleIndex = np.ndarray
ChunkedIndex = tuple[np.ndarray, np.ndarray]


def take(from_, by):
    match (from_, by):
        case (np.ndarray(), np.ndarray()):
            return __take_simple_from_simple(from_, by)
        case (np.ndarray(), (np.ndarray(), np.ndarray())):
            return __take_chunked_from_simple(from_, by)
        case ((np.ndarray(), np.ndarray()), np.ndarray()):
            return __take_simple_from_chunked(from_, by)
        case ((np.ndarray(), np.ndarray()), (np.ndarray(), np.ndarray())):
            return __take_chunked_from_chunked(from_, by)


def __take_simple_from_chunked(from_: ChunkedIndex, by: SimpleIndex):
    cumulative = np.insert(np.cumsum(from_[1]), 0, 0)[:-1]

    indices_into_chunks = np.argmax(by[:, np.newaxis] < cumulative, axis=1) - 1
    output = by - cumulative[indices_into_chunks] + from_[0][indices_into_chunks]
    return output


def __take_simple_from_simple(from_: np.ndarray, by: np.ndarray):
    return from_[by]


def __take_chunked_from_simple(from_: SimpleIndex, by: ChunkedIndex):
    output = np.zeros(by[1].sum(), dtype=int)
    output = __cfs_helper(from_, *by, output)
    return output


@nb.njit
def __cfs_helper(arr, starts, sizes, storage):
    rs = 0
    for i in range(len(starts)):
        cstart = starts[i]
        csize = sizes[i]
        storage[rs : rs + csize] = arr[cstart : cstart + csize]
        rs += csize
    return storage


@nb.njit
def __cfc_helper(from_starts, from_sizes, by_starts, by_sizes):
    pass


@nb.njit
def prefix_sum(arr):
    out = np.empty(len(arr) + 1, dtype=arr.dtype)
    total = 0
    out[0] = 0
    for i in range(len(arr)):
        total += arr[i]
        out[i + 1] = total
    return out


@nb.njit
def find_chunk(prefix, x):
    """
    Returns index i such that prefix[i] <= x < prefix[i+1].
    """
    lo = 0
    hi = len(prefix) - 1  # prefix has length N+1
    while lo + 1 < hi:
        mid = (lo + hi) // 2
        if prefix[mid] <= x:
            lo = mid
        else:
            hi = mid
    return lo


@nb.njit
def resolve_spanning_numba(
    start1, size1, start2, size2, out_start, out_size, out_owner
):
    """
    Resolves index2 slices into data-level chunks.
    Returns the number of output segments written.
    """
    prefix = prefix_sum(size1)
    out_pos = 0

    for j in range(len(start2)):
        logical = start2[j]
        remaining = size2[j]

        while remaining > 0:
            # Find which chunk in index1 we are inside
            i1 = find_chunk(prefix, logical)

            # Where inside that chunk?
            offset = logical - prefix[i1]

            # How many logical units remain in this chunk?
            chunk_left = size1[i1] - offset

            # How much we take
            take = chunk_left if chunk_left < remaining else remaining

            # Emit result
            out_start[out_pos] = start1[i1] + offset
            out_size[out_pos] = take
            out_owner[out_pos] = j

            out_pos += 1

            # Advance
            logical += take
            remaining -= take

    return out_pos


def __take_chunked_from_chunked(from_: ChunkedIndex, by: ChunkedIndex):
    if len(from_[0]) == 0 and from_[0][0] == 0:
        return by

    max_out = len(by[1]) * len(from_[1])
    out_start = np.empty(max_out, dtype=np.int64)
    out_size = np.empty(max_out, dtype=np.int64)
    out_owner = np.empty(max_out, dtype=np.int64)

    n = resolve_spanning_numba(
        from_[0], from_[1], by[0], by[1], out_start, out_size, out_owner
    )
    out_start = np.resize(out_start, (n,))
    out_size = np.resize(out_size, (n,))
    return out_start, out_size
